"""
<name>Build Subgroups</name>
<description>Runs subgroup discovery algorithms</description>
<icon>SubgroupDiscovery/SubgroupBuilder.png</icon>
<priority>10</priority>
"""

from OWWidget import *
import OWGUI
import sys
from SD_learner_classifier import *

class OWSubgroupBuilder(OWWidget):
    
    settingsList = [ "algorithm", "min_support", "g", "beam_width", "min_conf", "k", "max_rules"]
    def __init__(self, parent=None, signalManager = None, name='SubgroupBuilder'):
        OWWidget.__init__(self, parent, signalManager, name)
        
        self.inputs = [("Examples", ExampleTable, self.cdata)]
        self.outputs = [("Subgroup Descriptions", SDRules), ("Learner", orange.Learner),("Classifier", orange.Classifier) ]

        # GUI
        wbN = OWGUI.widgetBox(self.controlArea, "Settings")

        # Settings
        algorithms = ["SD", "SD-Preprocess", "Apriori-SD", "CN2-SD"]
        self.algorithm = algorithms[0]
        self.min_support = 5;
        self.g = 5
        self.beam_width = 20
        self.min_conf = 80
        self.k = 5
        self.name = "Subgroup SD"
        self.max_rules = 0

        self.loadSettings()

        OWGUI.lineEdit(wbN, self, "name", label="Name     ", orientation='horizontal', box=None, tooltip="Name", callback=None, valueType = str, validator=None)

        OWGUI.separator(self.controlArea)   
        
        OWGUI.comboBox(wbN, self, "algorithm", box=None, label="Algorithm", labelWidth=None, orientation='vertical', items=algorithms, tooltip="Choose an algorithem", sendSelectedValue = 1, callback = self.algorithmChanged)
        
        self.spin_support = OWGUI.spin(wbN, self, "min_support", 0, 100, 1, None, "Minimal support [%] ",labelWidth=200, orientation="horizontal", callback=self.settingChanged)
        self.spin_g       = OWGUI.spin(wbN, self, "g", 1, 1000, 1, None, "Generalization parameter ",labelWidth=200, orientation="horizontal",callback=self.settingChanged)
        self.spin_beam_width = OWGUI.spin(wbN, self, "beam_width", 3, 100, 1, None, "Beam width ",labelWidth=200, orientation="horizontal",callback=self.settingChanged)
        self.spin_conf    = OWGUI.spin(wbN, self, "min_conf", 0, 100, 1, None, "Minimal confidence [%]",labelWidth=200, orientation="horizontal",callback=self.settingChanged)
        self.spin_k       = OWGUI.spin(wbN, self, "k", 3, 100, 1, None, "k - num of times covered before removed ", labelWidth=200,orientation="horizontal",callback=self.settingChanged)
        self.spin_max_rules = OWGUI.spin(wbN, self, "max_rules", 0, 20, 1, None, "Max. number of subgroups (0=no limitations) ", labelWidth=200,orientation="horizontal",callback=self.settingChanged)
        self.algorithmChanged()

        OWGUI.separator(self.controlArea)

        #target Class
        self.lblTargetValue = QLabel("Target value:", wbN)
        self.listTargetValue = QListBox ( wbN, "listTargetValue" )
        self.listTargetValue.insertItem("No data")
        self.connect(self.listTargetValue, SIGNAL("selectionChanged()"), self.targetClassChanged)
        self.listTargetValue.blockSignals (true)  # block signals until trehe are no rules to send

        OWGUI.separator(wbN)

        OWGUI.button(wbN, self, "Build subgroups", callback=self.build)
        OWGUI.button(wbN, self, "Save ruleset to file in PMML format ...", callback=self.printToPmml)

        self.learner = SD_learner(name=self.name, algorithm=self.algorithm ,minSupport = self.min_support/100.0, minConfidence=self.min_conf/100.0 , beamWidth = self.beam_width, g = self.g, k=self.k, max_rules =self.max_rules)
        self.classifier = None
        self.subgroups = None

        # main area
        self.boxMainArea = QVBoxLayout(self.mainArea)

        self.edtRules = QMultiLineEdit(self.mainArea)                # we print the rules in this multi line edit
        self.edtRules.setReadOnly(TRUE)
        self.boxMainArea.addWidget(self.edtRules)

        self.resize(700,450)

    def algorithmChanged(self):
        """Set which setting are enabled and disabled for each algorithm and set learner."""
        if self.algorithm == "SD" or self.algorithm == "SD-Preprocess":
            self.spin_support.setDisabled(0)
            self.spin_g.setDisabled(0)      
            self.spin_beam_width.setDisabled(0)
            self.spin_conf.setDisabled(1)    
            self.spin_k.setDisabled(1)       
        elif self.algorithm == "Apriori-SD":
            self.spin_support.setDisabled(0)
            self.spin_g.setDisabled(1)      
            self.spin_beam_width.setDisabled(1)
            self.spin_conf.setDisabled(0)    
            self.spin_k.setDisabled(0)       
        else:            
            self.spin_support.setDisabled(1)
            self.spin_g.setDisabled(1)      
            self.spin_beam_width.setDisabled(1)
            self.spin_conf.setDisabled(1)    
            self.spin_k.setDisabled(0)
        self.settingChanged()

    def settingChanged(self):
        self.learner = SD_learner(name=self.name, algorithm=self.algorithm ,minSupport = self.min_support/100.0, minConfidence=self.min_conf/100.0 , beamWidth = self.beam_width, g = self.g, k=self.k, max_rules =self.max_rules)
        self.learner.name = self.name
        self.send("Learner", self.learner)

    def cdata(self, dataset):
        self.classifier = None
        self.subgroups = None
        self.send("Subgroup Descriptions", self.subgroups)
        self.send("Classifier", self.classifier)
        self.listTargetValue.clear()

        self.data = self.isDataWithClass(dataset, orange.VarTypes.Discrete) and dataset or None

        if self.data:
            #fill combo box with all possible target class values
            self.data = dataset
            valueObj= self.data.domain.classVar.firstvalue()
            while valueObj:
                self.listTargetValue.insertItem(valueObj.value)
                valueObj = self.data.domain.classVar.nextvalue(valueObj)
            self.listTargetValue.setCurrentItem(0)
        else:
            self.listTargetValue.insertItem('No data')
            self.listTargetValue.blockSignals (true)


    def build(self):        
        self.classifier = self.learner(self.data)
        self.edtRules.append(" %s %s supp=%d conf=%d bw=%0.2f g=%d k=%d rules=%d"%(self.name, self.algorithm , self.min_support, self.min_conf , self.beam_width,  self.g, self.k, self.max_rules))        
     
        # send classifier
    
        self.classifier.name = self.name
        self.send("Classifier", self.classifier)

        self.listTargetValue.blockSignals ( false)  # enable signals, enable sending rules to other widgets
        if self.listTargetValue.currentItem() <> -1:
            self.subgroups = self.classifier.getRules(targetClass= self.listTargetValue.currentText().ascii())
            self.subgroups.name = self.name
            self.send("Subgroup Descriptions", self.subgroups)

        # debug
        for rule in self.subgroups.rules:
            self.edtRules.append(rule.ruleToString())
        self.edtRules.append("----------------------------------")
    
    def printToPmml(self):
        if self.classifier == None:
            self.build()
        self.browseFile()
        outfile = open(self.filename, 'w')
        self.classifier.toPMML(outfile)
        outfile.close()
        
        
    def browseFile(self):
        tmpfilename = QFileDialog.getSaveFileName(
            "c:\ ",
            "Extensible Markup Language file (*.xml)",
            self,
            "open file dialog",
            "Enter a file" );
        
        tmpfilename = lower (str(tmpfilename))

        if tmpfilename.find(".") == -1:
            tmpfilename += ".xml"
        self.filename = tmpfilename

        #if (tmpfilename!=''):
        #    tmpfilename=str(tmpfilename)
        #    self.filename = tmpfilename.replace("/","\\\\")


    def targetClassChanged(self):
        if self.classifier:
            self.subgroups = self.classifier.getRules(targetClass= self.listTargetValue.currentText().ascii())
            self.edtRules.append(self.listTargetValue.currentText())
            self.subgroups.name = self.name
            self.send("Subgroup Descriptions", self.subgroups)

            # debug
            self.edtRules.append("target class changed")
            if self.subgroups:
                for rule in self.subgroups.rules:
                    self.edtRules.append(rule.ruleToString())
                self.edtRules.append("----------------------------------")


if __name__=="__main__":
    appl = QApplication(sys.argv)
    ow = OWSubgroupBuilder()
    appl.setMainWidget(ow)
    ow.show()

    filename = "..\\..\\doc\\datasets\\lenses.tab"
    if 'linux' in sys.platform:
        filename= "/usr/doc/orange/datasets/lenses.tab"
    dataset = orange.ExampleTable(filename)
    
    ow.cdata(dataset)
    ow.saveSettings()
    appl.exec_loop()

